// @author Zach Archer - http://zacharcher.com/

#include <Adafruit_NeoPixel.h>
#include "Const.h"
#include "Color.h"

#define randomf()                (random(0xffff)*(1.0/0xffff))
#define randomRangef(low,high)   (low + (random(0xffff)*(1.0/0xffff))*(high-low)) 
#define maybe()                  (random(0xf)&0x1)
#define lerp(a,b,prog)           ((a) + ((b)-(a)) * (prog))

const uint_fast8_t LED_STRIP_PIN = 4;
const uint_fast8_t BLINK_PIN = 13;

// LED layout 
const uint_fast16_t LED_COUNT = 57;
const uint_fast16_t TOP_COUNT = 29;
const uint_fast16_t BOTTOM_COUNT = LED_COUNT - TOP_COUNT;
const uint_fast8_t RGB_BUFFER_LEN = LED_COUNT * 3;

Adafruit_NeoPixel strip = Adafruit_NeoPixel( LED_COUNT, LED_STRIP_PIN, NEO_GRB + NEO_KHZ800 );

// Color randomness
const float MIN_S = 1.0;
const float MAX_S = 1.0;
// V values: Was 0.55. This seemed too dark; didn't illuminate the ground very well.
//           Trying 0.70 now. Hopefully this is a good compromise between safety (high
//           contrast animations are easy to spot), brightness, and battery life.
const float MIN_V = 0.70;
const float MAX_V = 0.70;
 
// Gradient animation
HSV hsv0 = (HSV){0,0,0};
ColorTween tween0;
HSV hsv1 = (HSV){0,0,0};
ColorTween tween1;
unsigned long tweenColorsAtMicros = 0;
const float COLOR_TWEEN_INTERVAL = 6.5;
const float COLOR_TWEEN_TIME = 2.0;

// Roadbumps
float bumpPhase = 0.0;
const float BUMP_INC = (0.0146) * 0.62;
const float BUMP_WIDTH_MULT = 0.2;  // larger == narrower bumps

//uint8_t test = 0;

void setup()
{
  pinMode(BLINK_PIN, OUTPUT);
  
  strip.begin();
  strip.show(); // Initialize all pixels to 'off'
}
 
void loop()
{
  //
  // ANIMATION: Colored gradients
  //
  if( micros() > tweenColorsAtMicros ) {
    // Don't use blue (Portland traffic code forbids blue light on the road, except
    // for emergency vehicles.) Select a distance to use between hues, then select
    // two hues in range magenta-RYG-cyan.
    float rando = randomf();  // [0..1]
    rando *= rando; // Still range [0..1], weighted towards magenta, because I like it :P
    
    float hueDistance = randomRangef( 0.5/6.0, 2.0/6.0 );
    float hue0 = rando * ((3.0/6.0) - hueDistance); // range is 2/3rds of color wheel: [red..cyan]
    hue0 -= (1.0/6.0);  // shift down to range: [magenta..green]

    // Random chance of hue0 or hue1 being higher
    float hue1 = hue0;
    if( maybe() ) {
      hue0 += hueDistance;
    } else {
      hue1 += hueDistance;
    }

    HSV new0 = (HSV){ hue0, randomRangef(MIN_S,MAX_S), randomRangef(MIN_V,MAX_V) };
    HSV new1 = (HSV){ hue1, randomRangef(MIN_S,MAX_S), randomRangef(MIN_V,MAX_V) };
    
    Color::startColorTweenUsingHSV( &tween0, hsv0, new0, COLOR_TWEEN_TIME );
    Color::startColorTweenUsingHSV( &tween1, hsv1, new1, COLOR_TWEEN_TIME );
    
    hsv0 = new0;
    hsv1 = new1;
    
    // After 5 seconds: Do another tween
    tweenColorsAtMicros = micros() + (unsigned long)(COLOR_TWEEN_INTERVAL*1000000);
  }
  
  // Prepare the color
  HSV end0 = hsv0;
  HSV end1 = hsv1;
  if( !tween0.isComplete ) {
    end0 = Color::colorTweenTickReturnHSV( &tween0 );
    end1 = Color::colorTweenTickReturnHSV( &tween1 );
  }
  
  // Advance the roadbumps. Trying to inject a little noise into the bump speed here;
  // hopefully will look a bit organic.
  bumpPhase += BUMP_INC * (0.75 + randomf()*0.5);
  if( bumpPhase >= 1.0 ) bumpPhase -= 1.0;  // Preserve floating-point precision :P

  float ratio = 0.0;
  for( uint8_t i=0; i<TOP_COUNT; i++ ) {
    ratio += (1.0/TOP_COUNT); // iterate [0..1]
    
    // Roadbump here?
    // Bump value is confusing! 1.0==NO bump. 0.0==FULL bump.
    // Why? Because bump value is used for saturation in HSV color creation,
    // so 0.0==white, appears like a bright bump. Ah, nice one.
    float bump = 1.0;
    
    float bumpRegion = bumpPhase + ratio*BUMP_WIDTH_MULT;
    bumpRegion -= floor(bumpRegion);
    
    if( bumpRegion < 0.25 ) {
      bump = 1.0 - max( 0.0, sin( bumpRegion * M_PI*16.0 ) );
    }
    
    HSV hsv = Color::lerpHSV( end0, end1, ratio );
    hsv.s *= bump;
    hsv.v = lerp( 1.0, hsv.v, bump ); // Bumps should have full brightness
    
    RGB rgb = Color::hsv2rgb( hsv.h, hsv.s, hsv.v );

    // Convert rgb floats to bytes, 0..0xff
    uint_fast8_t r = Color::gammaValue( rgb.r * 0xff );
    uint_fast8_t g = Color::gammaValue( rgb.g * 0xff );
    uint_fast8_t b = Color::gammaValue( rgb.b * 0xff );

    // 0x00ff00 sets LED to yellow (wtf argh). Moral of the story:
    // DON'T USE THIS FUNCTION SIGNATURE:  
    //       setPixelColor(uint16_t n, uint32_t c),
    // USE THIS ONE:                       
    //       setPixelColor(uint16_t n, uint8_t r, uint8_t g, uint8_t b),

    // Push to Top LED strip.
    strip.setPixelColor( i, r, g, b );

    // Push to Bottom LED strip. NeoPixel library ignores LEDs that are out of range,
    // so don't worry about whether LED index is a sane value or not.
    strip.setPixelColor( TOP_COUNT*2 - i, r, g, b );
  }
    
  //
  // TIMING (FIXME?)
  //
  
  /*
  // Pause between frames of animation
  while( micros() < nextFrameAtMicros );
  
  // Set the next frame time
  nextFrameAtMicros = micros() + microsPerFrame;
   */
    
  strip.show();
   
  // Blink the onboard LED to prove we're alive
  static uint_fast8_t pulse = 0x0;
  pulse++;
  digitalWrite( BLINK_PIN, (pulse&0x3f)==0 );
}
